<?php
/**
 * Revisionsrater plugin based on changes and rater plugins
 * https://www.dokuwiki.org/plugin:changes
 * https://www.dokuwiki.org/plugin:rater
 *
 * @license    GPL 2 (http://www.gnu.org/licenses/gpl.html)
 * @author     GEIST Research Group <geist@agh.edu.pl>
 */

// must be run within DokuWiki
if (!defined('DOKU_INC')) die();

/**
 * All DokuWiki plugins to extend the parser/rendering mechanism
 * need to inherit from this class
 */
class syntax_plugin_revisionsrater extends DokuWiki_Syntax_Plugin
{

    protected $raterConfig = array();

    public function getType()
    {
        return 'substition';
    }

    public function getSort()
    {
        return 32;
    }

    public function connectTo($mode)
    {
        $this->Lexer->addSpecialPattern('\{revisionsrater>[^}]*\}', $mode, 'plugin_revisionsrater');
    }

    public function handle($match, $state, $pos, Doku_Handler $handler)
    {
        $namespace = substr($match, 16, -1);
        //handle rater parameters
        return $namespace;
    }

    public function render($mode, Doku_Renderer $R, $namespace)
    {

        if ($mode == 'xhtml') {
            /* @var Doku_Renderer_xhtml $R */
            $R->info['cache'] = false;
            $changes = $this->getChanges($namespace);
            if (!count($changes)) return true;

            $this->renderSimpleList($changes, $R);
            return true;
        } elseif ($mode == 'metadata') {
            /* @var Doku_Renderer_metadata $R */
            global $conf;
            $R->meta['relation']['depends']['rendering'][$conf['changelog']] = true;
            return true;
        }
        return false;
    }

    /**
     * Based on getRecents() from inc/changelog.php
     *
     * @param int $num
     * @param array $ns
     * @param array $excludedpages
     * @param array $type
     * @param array $user
     * @param int $maxage
     * @return array
     */
    protected function getChanges($namespace)
    {
        global $conf;
        global $INFO;

        $changes = array();
        $lines = @file($conf['changelog']);

        // Merge media changelog
        if ($this->getConf('listmedia')) {
            $linesMedia = @file($conf['media_changelog']);
            // Add a tag to identiy the media lines
            foreach ($linesMedia as $key => $value) {
                $linesMedia[$key] = $value . "media";
            }
            $lines = array_merge($lines, $linesMedia);
        }


        for ($i = count($lines) - 1; $i >= 0; $i--) {
            $change = $this->handleChangelogLine($lines[$i], $namespace);
            if ($change !== false) {
                $changes[] = $change;
            }
        }

        // Date sort merged page and media changes
        if ($this->getConf('listmedia')) {
            $dates = array();
            foreach ($changes as $change) {
                $dates[] = $change['date'];
            }
            array_multisort($dates, SORT_ASC, $changes);
        }
        return $changes;
    }


    protected function handleChangelogLine($line, $namespace)
    {
        // split the line into parts
        $change = parseChangelogLine($line);
        if ($change === false) return false;

        // check if it's a hidden page
        if (isHiddenPage($change['id'])) return false;

        if (isset($namespace)) {
            if (!$this->isInNamespace($namespace, $change['id'])) return false;
        }

        // check ACL
        $change['perms'] = auth_quickaclcheck($change['id']);
        if ($change['perms'] < AUTH_READ) return false;

        return $change;
    }

    /**
     * Check if page belongs to one of namespaces in the list
     *
     * @param array $namespaces
     * @param string $id page id
     * @return bool
     */
    protected function isInNamespace($namespaces, $id)
    {
        if ($namespaces === $id) {
            return true;
        }
        return false;
    }

    /**
     * Render with a simple list render
     *
     * @param array $changes
     * @param Doku_Renderer_xhtml $R
     * @param null $flags
     */
    protected function renderSimpleList($changes, &$R)
    {
        global $conf;
        $flags = ['summary', 'signature', 'dayheaders'];
        $flags = $this->parseSimpleListFlags($flags);

        $dayheaders_date = '';
//        if($flags['dayheaders']) {
//            $dayheaders_date = date('Ymd', $changes[0]['date']);
//            $this->dayheader($R, $changes[0]['date']);
//        }

        $R->listu_open();
        foreach ($changes as $change) {
            $this->renderChangeRow($R, $change, $dayheaders_date, $flags);
            $this->renderRating($R, $change);
        }
        $R->listu_close();
    }

    protected function renderChangeRow($renderer, $change, $dayheaders_date, $flags)
    {
        global $conf;

        $R = $renderer;
        $linkToRevision = ':' . $change['id'];
        $linkName = $change['id'] . '_' . $change['date'];

//        $tdate = date('Ymd', $change['date']);
//        if($tdate != $dayheaders_date) {
//            $R->listu_close(); // break list to insert new header
//            $this->dayheader($R, $change['date']);
//            $R->listu_open();
//            $dayheaders_date = $tdate;
//        }

        $R->listitem_open(1);
        $R->listcontent_open();
        $linkxhtml = null;
        if (trim($change['extra']) == 'media') {
            $linkxhtml = $R->internalmedia($linkToRevision, $linkName, null, false, 'navigation');
        } else {
            $linkxhtml = $R->internallink($linkToRevision, $linkName, null, true, 'navigation');
        }
        $R->doc .= preg_replace('/(href=".*?)((#.*?)?")/', '$1&rev=' . $change['date'] . '$2', $linkxhtml);


        if ($flags['summary']) {
            $R->cdata(' ' . $change['sum']);
        }
        if ($flags['signature']) {
            $userRealName = $this->getUserName($change);
            $date = strftime($conf['dformat'], $change['date']);
            $R->cdata(' ');
            $R->entity('---');
            $R->cdata(' ');
            $R->cdata($change['user']);
            $R->emphasis_open();
            $R->cdata(' (');
            $R->cdata($userRealName);
            $R->emphasis_close();
            $R->cdata(')');
            $R->cdata(' ' . $date);
        }
        $R->listcontent_close();
        $R->listitem_close();
    }

    protected function renderRating($renderer, $change)
    {
        global $ID;
        global $lang;
        global $conf;
        global $INFO;


        $ret = '';
        $renderer->info['cache'] = false;

//        $this->getRaterConfiguration();

        // Config settings
        $rater_ip_voting_restriction = $this->getConf('voting_restriction'); // restrict ip address voting (true or false)
        $rater_ip_vote_qty = $this->getConf('vote_qty');           // how many times an ip address / user name can vote
        $rater_already_rated_msg = sprintf($this->getConf('already_rated_msg'), $rater_ip_vote_qty);
        $rater_not_selected_msg = $this->getConf('not_selected_msg');
        $rater_thankyou_msg = $this->getConf('thankyou_msg');
        $rater_generic_text = $this->getConf('generic_text');       // generic item text
        $rater_end_of_line_char = $this->getConf('eol_char');           // to separate the records
        if ($rater_end_of_line_char == '') $rater_end_of_line_char = '\n';

        $btn_submit = 'Rate';

//            $isLoggedUser = isset($INFO['userinfo']);
        $isLoggedUser = true;

//            $rater_id                    = trim($data['rater_id']);
//            $rater_name                  = trim($data['rater_name']);
//            $rater_headline              = trim($data['rater_headline']);
//            $rater_stat_sort             = trim($data['rater_stat_sort']);
//            $rater_img                   = $data['rater_img'];
//            $rater_zoom                  = $data['rater_zoom'];
//            $rater_sid                   = $data['rater_sid'];
//            $rater_ns                    = $data['rater_ns'];

        $rater_id = $change['id'] . '_' . $change['date'];
        $rater_headline = 'on';
        $rater_name = $change['id'] . '_' . $change['date'];

        $rateDetails = array(
            'page' => $change['id'],
            'revision' => $change['date'],
            'namespace' => false,
            'rate' => false,
        );

        if ($rateDetails['userinfo']) {
            $rateDetails['login'] = $INFO['client'];
        }

        if (!isset($data['rater_end'])) {
            $data['rater_end'] = 'never';
            $rater_end = 'never';
        } else {
            $rater_end = $data['rater_end'];
        }

        if (!isset($data['rater_tracedetails'])) $data['rater_tracedetails'] = '0';
        if (!isset($rater_id)) $rater_id = 1;
        $rater_item_name = $data['rater_name'];                              // item name what is to be rated
        if (!isset($rater_item_name)) $rater_item_name = $rater_generic_text;
//            $rater_type = $data['rater_type'];
        $rater_type = 'rate';
        if (!isset($rater_type)) $rater_type = "stars";

        // DO NOT MODIFY BELOW THIS LINE
        $rater_filename = metaFN($rater_id, '.rating');
        $rater_rating = 0;
        $rater_stars = "";
        $rater_stars_txt = "";
        $rater_rating = 0;
        $rater_votes = 0;
        $rater_msg = "";
        $today = date('d.m.Y');
        // check to trace ip or user name
        $user_grp = pageinfo();
        $rater_realname = $user_grp['userinfo']['name'];
        if ($data['rater_trace'] === 'user') {
            $rater_ip = $user_grp['userinfo']['name'];
        } else {
            $rater_ip = getenv("REMOTE_ADDR");
        }

        /******************************************************************************/
        // Rating action
        if (isset($_REQUEST["rate" . $rater_id])) {
            if (isset($_REQUEST["rating_" . $rater_id]) && $isLoggedUser) {
                while (list($key, $val) = each($_REQUEST["rating_" . $rater_id])) {
                    $rater_rating = $val;
                }
                $rateDetails['rate'] = $rater_rating;
                $rater_file = fopen($rater_filename, "a+");
                $rater_str = "";
                $rater_str = rtrim(fread($rater_file, 1024 * 8), $rater_end_of_line_char);
                if ($rater_str != "") {
                    if ($rater_ip_voting_restriction) {
                        $rater_data = explode($rater_end_of_line_char, $rater_str);
                        $rater_ip_vote_count = 0;
                        foreach ($rater_data as $d) {
                            $rater_tmp = explode("|", $d);
                            $rater_oldip = str_replace($rater_end_of_line_char, "", $rater_tmp[1]);
                            if ($rater_ip == $rater_oldip) {
                                $rater_ip_vote_count++;
                            }
                        }
                        if ($rater_ip_vote_count > ($rater_ip_vote_qty - 1)) {
                            $rater_msg = $rater_already_rated_msg;
                        } else {
                            fwrite($rater_file, $rater_rating . "|" . $rater_ip . $rater_end_of_line_char);
                            $rater_msg = $rater_thankyou_msg;
                        }
                    } else {
                        fwrite($rater_file, $rater_rating . "|" . $rater_ip . $rater_end_of_line_char);
                        $rater_msg = $rater_thankyou_msg;
                    }
                } else {
                    fwrite($rater_file, $rater_rating . "|" . $rater_ip . $rater_end_of_line_char);
                    $rater_msg = $rater_thankyou_msg;
                }
                fclose($rater_file);

                $this->writeToPROVFile($rateDetails);
            } else {
                $rater_msg = $rater_not_selected_msg;
            }

        }

        /******************************************************************************/

        // Get current rating
        $r1 = '0';
        $r2 = '0';
        $r3 = '0';
        $r4 = '0';
        $r5 = '0';
        $rater_votes = '0';
        $rater_sum = '0';

        if (is_file($rater_filename)) {
            $tmp_array = $this->calc_rater_rating($rater_filename);
            $rater_rating = $tmp_array[0][0];
            $r1 += $tmp_array[0][1];
            $r2 += $tmp_array[0][2];
            $r3 += $tmp_array[0][3];
            $r4 += $tmp_array[0][4];
            $r5 += $tmp_array[0][5];
            $rater_votes += $tmp_array[0][6];
            $rater_sum += $tmp_array[0][7];
        } else {
            $rater_file = fopen($rater_filename, "w");
            fclose($rater_file);
        }

        // Assign star image
        $rater_stars = $this->assign_star_image($rater_rating);
        $anker_id = 'rateanker_' . uniqid((double)microtime() * 1000000, 1);

        // build the return value for details
        $ret_details = '<p class="rating__details">';
        $alink_Details = '';

        if ($data['rater_tracedetails'] === '1') {
            if ($alink_Details === '') {
                $alink_Details = '<a href="#' . $anker_id . '" onclick="hidden' . $rater_id . '()">(Details)</a>';
            }
            $ret_details .= '<img src="' . DOKU_BASE . 'lib/plugins/revisionsrater/img/1star.gif?w=40&amp;" alt="1 Star" width="40" align="left" /> ' . $r1 . ' visitor votes<br />';
            $ret_details .= '<img src="' . DOKU_BASE . 'lib/plugins/revisionsrater/img/2star.gif?w=40&amp;" alt="2 Stars" width="40" align="left" /> ' . $r2 . ' visitor votes<br />';
            $ret_details .= '<img src="' . DOKU_BASE . 'lib/plugins/revisionsrater/img/3star.gif?w=40&amp;" alt="3 Stars" width="40" align="left" /> ' . $r3 . ' visitor votes<br />';
            $ret_details .= '<img src="' . DOKU_BASE . 'lib/plugins/revisionsrater/img/4star.gif?w=40&amp;" alt="4 Stars" width="40" align="left" /> ' . $r4 . ' visitor votes<br />';
            $ret_details .= '<img src="' . DOKU_BASE . 'lib/plugins/revisionsrater/img/5star.gif?w=40&amp;" alt="5 Stars" width="40" align="left" /> ' . $r5 . ' visitor votes';
        }
        $ret_details .= '</p>';


        // Output rate
        $rater_img_xhtml = "";
//                if($rater_img!=false) {
//                    // render the image link and output centralized within rater box
//                    $info = array();
//                    $rater_img_xhtml = p_render('xhtml',p_get_instructions("{{".$rater_img."}}"),$info);
//                    $rater_img_xhtml = str_replace("<p>","",$rater_img_xhtml);
//                    $rater_img_xhtml = str_replace("</p>","",$rater_img_xhtml);
//                    $rater_img_xhtml  = '<tr><td class="rater_item">'.$rater_img_xhtml.'</td></tr>'.NL;
//                }

        $ret .= '<form method="post" action="doku.php?id=' . $ID . '" >
                       <table class="rater_table">' . NL;
        $ret .= $rater_img_xhtml;
        $ul_rater_item_name = str_replace("_", " ", $rater_item_name);
        if ($rater_headline !== "off") {
            $ret .= '<tr>
                           <td class="rater_item">' . $ul_rater_item_name . '</td>
                         </tr>' . NL;
        }
        $ret .= '<tr>
                        <td class="rater_rating" id="' . $anker_id . '">
                          <span><script type="text/javascript" language="JavaScript1.2">
                            var visible = false;
                            function hidden' . $rater_id . '() 
                            {   if (visible)
                                {   document.getElementById("details_' . $rater_id . '").style.display = "none";
                                    visible = false; }
                                else
                                {   document.getElementById("details_' . $rater_id . '").style.display = "block";
                                    visible = true; }
                            } 
                          </script></span>
                        <div style="display:inline;margin-left:auto; margin-right:0;">
                        <div style="display:table-cell; text-align:left;">
                        <img src="' . $rater_stars . '?x=' . uniqid((double)microtime() * 1000000, 1) . '" alt="' . $rater_stars_txt . ' stars" />&nbsp;' . $rater_stars_txt . ' from 
                        <span class="reviewcount"> ' . $rater_votes . ' votes ' . $alink_Details . '</span><br />';

        if ($isLoggedUser) {
            $ret .= '<label for="rate1_' . $rater_id . '"><input type="radio" value="1" name="rating_' . $rater_id . '[]" id="rate1_' . $rater_id . '" />&nbsp;</label>';
            $ret .= '<label for="rate2_' . $rater_id . '"><input type="radio" value="2" name="rating_' . $rater_id . '[]" id="rate2_' . $rater_id . '" />&nbsp;</label>';
            $ret .= '<label for="rate3_' . $rater_id . '"><input type="radio" value="3" name="rating_' . $rater_id . '[]" id="rate3_' . $rater_id . '" />&nbsp;</label>';
            $ret .= '<label for="rate4_' . $rater_id . '"><input type="radio" value="4" name="rating_' . $rater_id . '[]" id="rate4_' . $rater_id . '" />&nbsp;</label>';
            $ret .= '<label for="rate5_' . $rater_id . '"><input type="radio" value="5" name="rating_' . $rater_id . '[]" id="rate5_' . $rater_id . '" />&nbsp;</label>';
            $ret .= '<input type="hidden" name="rs_id" value="' . $rater_id . '" />';
            $ret .= '<input type="hidden" name="rs_anker" value="' . $anker_id . '" />';
            $ret .= '<input type="submit" name="rate' . $rater_id . '" value="' . $btn_submit . '" />'; // </td></tr>';
        }

        $ret .= '<div style="display : none" id="details_' . $rater_id . '">' . $ret_details . '</div> </div></div></td></tr>';

        if ($rater_msg != "") $ret .= "<div>" . $rater_msg . "</div>";
        $ret .= '</table>';
        $ret .= '</form>';
        // Render
        $renderer->doc .= $ret;
    }

    /**
     * Render the day header
     *
     * @param Doku_Renderer $R
     * @param int $date
     */
    protected function dayheader(&$R, $date)
    {
        if ($R->getFormat() == 'xhtml') {
            /* @var Doku_Renderer_xhtml $R */
            $R->doc .= '<h3 class="changes">';
            $R->cdata(dformat($date, $this->getConf('dayheaderfmt')));
            $R->doc .= '</h3>';
        } else {
            $R->header(dformat($date, $this->getConf('dayheaderfmt')), 3, 0);
        }
    }

    /**
     * Parse flags for the simple list render
     *
     * @param array $flags
     * @return array
     */
    protected function parseSimpleListFlags($flags)
    {
        $outFlags = array('summary' => true, 'signature' => false, 'dayheaders' => false);
        if (!empty($flags)) {
            foreach ($flags as $flag) {
                if (array_key_exists($flag, $outFlags)) {
                    $outFlags[$flag] = true;
                } elseif (substr($flag, 0, 2) == 'no') {
                    $flag = substr($flag, 2);
                    if (array_key_exists($flag, $outFlags)) {
                        $outFlags[$flag] = false;
                    }
                }
            }
        }
        return $outFlags;
    }

    /**
     * Get username or fallback to ip
     *
     * @param array $change
     * @return mixed
     */
    protected function getUserName($change)
    {
        /* @var DokuWiki_Auth_Plugin $auth */
        global $auth;
        if (!empty($change['user'])) {
            $user = $auth->getUserData($change['user']);
            return $user['name'];
        } else {
            return $change['ip'];
        }
    }

    protected function assign_star_image($rater_rating)
    {
        // Assign star image
        if ($rater_rating <= 0) {
            $rater_stars = DOKU_BASE . "lib/plugins/revisionsrater/img/00star.gif";
            $rater_stars_txt = "Not Rated";
        }
        if ($rater_rating >= 0.5) {
            $rater_stars = DOKU_BASE . "lib/plugins/revisionsrater/img/05star.gif";
            $rater_stars_txt = "0.5";
        }
        if ($rater_rating >= 1) {
            $rater_stars = DOKU_BASE . "lib/plugins/revisionsrater/img/1star.gif";
            $rater_stars_txt = "1";
        }
        if ($rater_rating >= 1.5) {
            $rater_stars = DOKU_BASE . "lib/plugins/revisionsrater/img/15star.gif";
            $rater_stars_txt = "1.5";
        }
        if ($rater_rating >= 2) {
            $rater_stars = DOKU_BASE . "lib/plugins/revisionsrater/img/2star.gif";
            $rater_stars_txt = "2";
        }
        if ($rater_rating >= 2.5) {
            $rater_stars = DOKU_BASE . "lib/plugins/revisionsrater/img/25star.gif";
            $rater_stars_txt = "2.5";
        }
        if ($rater_rating >= 3) {
            $rater_stars = DOKU_BASE . "lib/plugins/revisionsrater/img/3star.gif";
            $rater_stars_txt = "3";
        }
        if ($rater_rating >= 3.5) {
            $rater_stars = DOKU_BASE . "lib/plugins/revisionsrater/img/35star.gif";
            $rater_stars_txt = "3.5";
        }
        if ($rater_rating >= 4) {
            $rater_stars = DOKU_BASE . "lib/plugins/revisionsrater/img/4star.gif";
            $rater_stars_txt = "4";
        }
        if ($rater_rating >= 4.5) {
            $rater_stars = DOKU_BASE . "lib/plugins/revisionsrater/img/45star.gif";
            $rater_stars_txt = "4.5";
        }
        if ($rater_rating >= 5) {
            $rater_stars = DOKU_BASE . "lib/plugins/revisionsrater/img/5star.gif";
            $rater_stars_txt = "5";
        }
        return $rater_stars;
    }

    protected function calc_rater_rating($rater_filename)
    {
        $rater_end_of_line_char = $this->getConf('eol_char');
        if ($rater_end_of_line_char == '') $rater_end_of_line_char = '\n';
        $rater_file = fopen($rater_filename, "r");
        $rater_str = "";
        $rater_str = fread($rater_file, 1024 * 8);
        if ($rater_str != "") {
            $rater_data = explode($rater_end_of_line_char, $rater_str);
            $rater_votes = count($rater_data) - 1;
            $rater_sum = 0;

            foreach ($rater_data as $d) {
                $d = explode("|", $d);
                $rater_sum += $d[0];

                // collect votes per level to display the details
                if ($d[0] === '1') {
                    $r1 = $r1 + 1;
                }
                if ($d[0] === '2') {
                    $r2 = $r2 + 1;
                }
                if ($d[0] === '3') {
                    $r3 = $r3 + 1;
                }
                if ($d[0] === '4') {
                    $r4 = $r4 + 1;
                }
                if ($d[0] === '5') {
                    $r5 = $r5 + 1;
                }
            }
            if ($rater_votes > 0) $rater_rating = number_format(($rater_sum / $rater_votes), 2, '.', '');
        }
        fclose($rater_file);
        $tmp_array[] = array($rater_rating, $r1, $r2, $r3, $r4, $r5, $rater_votes, $rater_sum);
        return $tmp_array;
    }

    protected function writeToPROVFile($rateDetails)
    {
        if (!plugin_isdisabled('prov')) {
            $prov =& plugin_load('action', 'prov_prov');
            $prov->saveRevisionRate($rateDetails);
        }
    }
}

